package me.hammerle.code;

import java.awt.Color;
import me.hammerle.exceptions.HoldCodeException;
import me.hammerle.exceptions.IllegalStringException;
import me.hammerle.exceptions.NoSuchMethodException;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.time.ZonedDateTime;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Collections;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.function.BiFunction;
import java.util.function.BiConsumer;
import java.util.stream.Collectors;
import me.hammerle.console.ConsoleUtils;
import me.hammerle.graphics.*;
import me.hammerle.scheduler.SnuviScheduler;

public class CodeParser 
{
    private final static HashMap<String, BiFunction<Object[], Script, Object>> METHODS = new HashMap<>();
    
    private static void registerFunction(String s, BiFunction<Object[], Script, Object> f)
    {
        METHODS.put(s, f);
    }
    
    private static void registerConsumer(String s, BiConsumer<Object[], Script> f)
    {
        METHODS.put(s, (BiFunction<Object[], Script, Object>) (args, sc) -> 
        {
            f.accept(args, sc);
            return Void.TYPE;
        });
    }
    
    private static void registerAlias(String s, String original)
    {
        registerFunction(s, METHODS.get(original)); 
    }
    
    @SuppressWarnings("unchecked")
    public static void initCodeParser()
    {
        registerConsumer("test", (args, sc) -> 
                test(args, sc));
        registerFunction("nothing", (args, sc) -> 
                0);
        registerFunction("", (args, sc) -> 
                args[0]);
        registerConsumer("error", (args, sc) -> 
                printStack = !printStack);
        
        // -------------------------------------------------------------    
        // Script-Consolen-Bibliothek 
        // -------------------------------------------------------------  
        registerConsumer("script.list", (args, sc) -> 
                { ScriptUtils.print("The following scripts are active:"); ScriptControl.getScripts().forEach(s -> ScriptUtils.print(s.getId() + " | " + s.getName()));}); 
        registerConsumer("script.load", (args, sc) -> 
                ScriptControl.startScriptFromFile(args[0].toString())); 
        registerConsumer("clear", (args, sc) -> 
                ConsoleUtils.clear()); 
        registerConsumer("script.term", (args, sc) -> 
                termScript(args, sc)); 
        
        // -------------------------------------------------------------    
        // Event-Bibliothek 
        // -------------------------------------------------------------  
        registerConsumer("event.load", (args, sc) -> 
                sc.loadEvent(args[0].toString())); 
        registerConsumer("event.unload", (args, sc) -> 
                sc.unloadEvent(args[0].toString())); 
        registerFunction("event.isloaded", (args, sc) -> 
                sc.isLoadedEvent(args[0].toString())); 

        // -------------------------------------------------------------    
        // Mathe-Bibliothek 
        // -------------------------------------------------------------    
        registerFunction("math.mod", (args, sc) -> 
                ((double) args[0]) % ((double) args[1]));
        registerFunction("math.abs", (args, sc) -> 
                Math.abs((double) args[0]));
        registerFunction("math.pow", (args, sc) -> 
                Math.pow((double) args[0], (double) args[1]));
        registerFunction("math.root", (args, sc) -> 
                Math.pow((double) args[0], 1d / ((double) args[1])));
        registerFunction("math.sin", (args, sc) -> 
                Math.sin((double) args[0]));
        registerFunction("math.cos", (args, sc) -> 
                Math.cos((double) args[0]));
        registerFunction("math.tan", (args, sc) -> 
                Math.tan((double) args[0]));
        registerFunction("math.asin", (args, sc) -> 
                Math.asin((double) args[0]));
        registerFunction("math.acos", (args, sc) -> 
                Math.acos((double) args[0]));
        registerFunction("math.atan", (args, sc) -> 
                Math.atan((double) args[0]));
        registerFunction("math.e", (args, sc) -> 
                Math.E);
        registerFunction("math.pi", (args, sc) -> 
                Math.PI);
        registerFunction("math.ln", (args, sc) -> 
                Math.log((double) args[0]));
        registerFunction("math.log", (args, sc) -> 
                Math.log10((double) args[0]));
        registerFunction("math.random", (args, sc) -> 
                (double) ScriptUtils.randomInt(getInt(args[0]), getInt(args[1])));
        registerFunction("math.round", (args, sc) -> 
                Math.round((double) args[0]));
        registerFunction("math.rounddown", (args, sc) -> 
                Math.floor((double) args[0]));
        registerFunction("math.roundup", (args, sc) -> 
                Math.ceil((double) args[0]));
        registerFunction("math.roundcomma", (args, sc) -> 
                new BigDecimal(((double) args[0])).setScale(getInt(args[1]), RoundingMode.HALF_UP).doubleValue());
        
        // -------------------------------------------------------------  
        // Listen-Bibliothek   
        // -------------------------------------------------------------    
        registerConsumer("list.new", (args, sc) ->                                                
                sc.setVar(args[0].toString(), new ArrayList<>()));
        registerFunction("list.exists", (args, sc) ->                                                
                args[0] instanceof List); 
        registerConsumer("list.add", (args, sc) ->                            
                ((List) args[0]).add(args[1]));
        registerConsumer("list.remove", (args, sc) ->                            
                ((List) args[0]).remove(args[1]));
        registerConsumer("list.removeindex", (args, sc) ->                            
                ((List) args[0]).remove(getInt(args[1])));                
        registerFunction("list.contains", (args, sc) ->                            
                ((List) args[0]).contains(args[1]));
        registerFunction("list.getsize", (args, sc) ->                            
                ((List) args[0]).size());
        registerFunction("list.getindex", (args, sc) ->                            
                ((List) args[0]).get(getInt(args[1])));
        registerConsumer("list.setindex", (args, sc) ->                            
                ((List) args[0]).set(getInt(args[1]), args[2]));
        registerConsumer("list.clear", (args, sc) ->                            
                ((List) args[0]).clear());
        registerFunction("list.getindexof", (args, sc) ->                            
                ((List) args[0]).indexOf(args[1]));  
        registerConsumer("list.sort", (args, sc) ->  
                sortList((List<Object>) args[0], sc));
        registerConsumer("list.reverse", (args, sc) ->                            
                Collections.reverse((List<Object>) args[0])); 
        registerConsumer("list.shuffle", (args, sc) ->                            
                Collections.shuffle((List<Object>) args[0]));

        // -------------------------------------------------------------  
        // Map-Bibliothek   
        // ------------------------------------------------------------- 
        registerConsumer("map.new", (args, sc) ->                                                
                sc.setVar(args[0].toString(), new HashMap<>())); 
        registerFunction("map.exists", (args, sc) ->                                                
                args[0] instanceof HashMap);  
        registerConsumer("map.add", (args, sc) ->                            
                ((HashMap) args[0]).put(args[1], args[2]));
        registerConsumer("map.remove", (args, sc) ->                            
                ((HashMap) args[0]).remove(args[1]));
        registerFunction("map.contains", (args, sc) ->                            
                ((HashMap) args[0]).containsKey(args[1]));
        registerFunction("map.getsize", (args, sc) ->                            
                ((HashMap) args[0]).size());
        registerFunction("map.get", (args, sc) ->                            
                ((HashMap) args[0]).get(args[1]));
        registerConsumer("map.clear", (args, sc) ->                            
                ((HashMap) args[0]).clear());
        
        // -------------------------------------------------------------  
        // Set-Bibliothek   
        // ------------------------------------------------------------- 
        registerConsumer("set.new", (args, sc) ->                                                
                sc.setVar(args[0].toString(), new HashSet<>())); 
        registerFunction("set.exists", (args, sc) ->                                                
                args[0] instanceof HashSet);  
        registerConsumer("set.add", (args, sc) ->                            
                ((HashSet) args[0]).add(args[1]));
        registerConsumer("set.remove", (args, sc) ->                            
                ((HashSet) args[0]).remove(args[1]));
        registerFunction("set.contains", (args, sc) ->                            
                ((HashSet) args[0]).contains(args[1]));
        registerFunction("set.getsize", (args, sc) ->                            
                ((HashSet) args[0]).size());

        // -------------------------------------------------------------  
        // Time-Bibliothek   
        // -------------------------------------------------------------
        registerFunction("time.get", (args, sc) ->                            
                (double) System.currentTimeMillis());
        registerFunction("time.nextday", (args, sc) ->         
                (double) getNextDay(args));   
        registerFunction("time.getyear", (args, sc) ->         
                (double) getYear(args));   
        registerFunction("time.getmonth", (args, sc) ->         
                (double) getMonth(args));   
        registerFunction("time.getday", (args, sc) ->         
                (double) getDay(args));   
        registerFunction("time.gethour", (args, sc) ->         
                (double) getHour(args));   
        registerFunction("time.getminute", (args, sc) ->         
                (double) getMinute(args));   
        registerFunction("time.getsecond", (args, sc) ->         
                (double) getSecond(args));                   

        // -------------------------------------------------------------    
        // Graphic-Bibliothek 
        // -------------------------------------------------------------
        registerConsumer("g.new", (args, sc) ->                            
                sc.newFrame(getInt(args[0]), getInt(args[1])));
        registerFunction("g.get", (args, sc) ->                            
                sc.getGraphicPanel().getGraphic(getInt(args[0])));
        registerConsumer("g.repaint", (args, sc) ->                            
                sc.getGraphicPanel().repaint());
        registerConsumer("g.newpoint", (args, sc) ->                            
                newPoint(args, sc));
        registerConsumer("g.newline", (args, sc) ->                            
                newLine(args, sc));
        registerConsumer("g.newoval", (args, sc) ->                            
                newOval(args, sc));
        registerConsumer("g.newrectangle", (args, sc) ->                            
                newRectangle(args, sc));
        registerConsumer("g.translate", (args, sc) ->                            
                ((Graphic) args[0]).translate(getInt(args[1]), getInt(args[2])));
        registerConsumer("g.setcolor", (args, sc) ->                            
                ((Colorable) args[0]).setColor((Color) args[1]));
        registerConsumer("g.setfilled", (args, sc) ->                            
                ((Colorable) args[0]).setFilled((boolean) args[1]));
        registerConsumer("g.setx1", (args, sc) ->                            
                ((Line) args[0]).setFirstX(getInt(args[1])));
        registerConsumer("g.setx2", (args, sc) ->                            
                ((Line) args[0]).setSecondX(getInt(args[1])));
        registerConsumer("g.sety1", (args, sc) ->                            
                ((Line) args[0]).setFirstY(getInt(args[1])));
        registerConsumer("g.sety2", (args, sc) ->                            
                ((Line) args[0]).setSecondY(getInt(args[1])));
        
        // -------------------------------------------------------------    
        // Read-Bibliothek
        // -------------------------------------------------------------  
        registerFunction("read.color", (args, sc) ->         
                getColor(args, sc));
        
        // -------------------------------------------------------------    
        // Ohne Bibliothek
        // -------------------------------------------------------------    
        registerFunction("add", (args, sc) -> 
                Arrays.stream(args).mapToDouble(s -> (double) s).sum());
        registerFunction("sub", (args, sc) -> 
                ((double) args[0]) - ((double) args[1]));
        registerFunction("inc", (args, sc) -> 
                increaseVar(args[0], sc, 1)); 
        registerFunction("dec", (args, sc) -> 
                increaseVar(args[0], sc, -1));
        registerFunction("mul", (args, sc) -> 
                ((double) args[0]) * ((double) args[1]));
        registerFunction("div", (args, sc) -> 
                ((double) args[0]) / ((double) args[1]));
        registerFunction("getvar", (args, sc) -> 
                sc.getVar(args[0].toString()));
        registerConsumer("setvar", (args, sc) -> 
                sc.setVar(args[0].toString(), args[1]));
        registerConsumer("removevar", (args, sc) -> 
                sc.removeVar(args[0].toString()));
        registerConsumer("debug", (args, sc) -> 
                ScriptUtils.printDebug(ScriptUtils.connect(args, 0)));
        registerConsumer("print", (args, sc) -> 
                print(args, sc));
        registerConsumer("reset", (args, sc) -> 
                sc.resetLoopCounter());
        registerConsumer("wait", (args, sc) -> 
                { sc.resetLoopCounter(); throw new HoldCodeException(); });
        registerConsumer("goto", (args, sc) -> 
                sc.gotoLabel(args[0].toString()));
        registerConsumer("gotoline", (args, sc) -> 
                { sc.gotoSpecialJumpLine(); sc.incLoopCounter(); });
        registerConsumer("sgoto", (args, sc) -> 
                scheduleGoto(args, sc));
        registerConsumer("gosub", (args, sc) -> 
                sc.gotoLabelWithReturn(args[0].toString()));
        registerConsumer("return", (args, sc) -> 
                sc.doReturn());
        registerConsumer("try", (args, sc) -> 
                sc.saveTryJumpLine());                   
        registerConsumer("catch", (args, sc) -> 
                sc.gotoSpecialJumpLine());
        registerConsumer("if", (args, sc) -> 
                ifFunction(args, sc));   
        registerConsumer("else", (args, sc) -> 
                sc.gotoSpecialJumpLine());   
        registerConsumer("while", (args, sc) -> 
                whileFunction(args, sc)); 
        registerFunction("equal", (args, sc) -> 
                isEqual(args));
        registerAlias("equals", "equal");
        registerFunction("less", (args, sc) -> 
                (double) args[0] < (double) args[1]);
        registerFunction("greater", (args, sc) -> 
                (double) args[0] > (double) args[1]);
        registerFunction("notequal", (args, sc) -> 
                !isEqual(args));
        registerFunction("lessequal", (args, sc) -> 
                (double) args[0] <= (double) args[1]);
        registerFunction("greaterequal", (args, sc) -> 
                (double) args[0] >= (double) args[1]);
        registerFunction("invert", (args, sc) -> 
                !((boolean) args[0]));
        registerFunction("and", (args, sc) -> 
                Arrays.stream(args).allMatch(s -> s.equals(true)));
        registerFunction("or", (args, sc) -> 
                Arrays.stream(args).anyMatch(s -> s.equals(true)));
        registerFunction("concatlist", (args, sc) ->      
                ((List<Object>) args[0]).stream().limit(getInt(args[3]) + 1).skip(getInt(args[2])).map(o -> o.toString()).collect(Collectors.joining(args[1].toString())));
        registerConsumer("split", (args, sc) ->      
                split(args, sc));             
        registerFunction("concat", (args, sc) ->                            
                ScriptUtils.connect(args, 0));
        registerFunction("tolowercase", (args, sc) ->                            
                ScriptUtils.connect(args, 0).toLowerCase());
        registerFunction("touppercase", (args, sc) ->                            
                ScriptUtils.connect(args, 0).toUpperCase());   
        registerConsumer("waitfor", (args, sc) ->     
                waitFor(args, sc));
    }
    
    public static boolean printStack = false;

    @SuppressWarnings("unchecked")
    public static Object parseFunction(Script sc, String function, Object[] args) throws HoldCodeException
    {
        try
        {
            BiFunction<Object[], Script, Object> f = METHODS.get(function);
            if(f == null)
            {
                throw new NoSuchMethodException(function);
            }
            return METHODS.get(function).apply(args, sc);
        }
        catch(Exception ex)
        {
            if(ex instanceof HoldCodeException)
            {
                throw ex;
            }
            if(printStack)
            {
                ex.printStackTrace();
            }
            if(sc.getTryJumpLine() != -1)
            {
                sc.setVar("error", ex.getClass().getSimpleName());
                sc.gotoTryJumpLine();
                sc.resetTryJumpLine();
                return Void.TYPE;
            }
            printQuestException(sc, ex, function + "(" + Arrays.stream(args).map(o -> String.valueOf(o)).collect(Collectors.joining(", ")) + ")");  
            throw new HoldCodeException();
        }
    }
    
    public static void printQuestException(Script sc, Exception ex, String line)
    {
        sc.resetLoopCounter();
        ScriptUtils.printError("Error in");
        ScriptUtils.printErrorList("script:", sc.getName());
        ScriptUtils.printErrorList("line:", line);
        ScriptUtils.printErrorList("line number:", sc.getActiveCodeLine());
        if(ex.getLocalizedMessage() == null)
        {
            ScriptUtils.printErrorList("exception:", ex.getClass().getSimpleName());
        }
        else
        {
            ScriptUtils.printErrorList("exception:", ex.getClass().getSimpleName() + " - " + ex.getLocalizedMessage());
        }
        if(ex instanceof IllegalStringException)
        {
            ScriptUtils.printErrorList("illegal value:", ((IllegalStringException) ex).getBadString());
        }
    }
    
    // -----------------------------------------------------------------------------------
    // Functions
    // -----------------------------------------------------------------------------------
    
    private static void termScript(Object[] args, Script sc)
    {
        if(args.length == 0)
        {
            ScriptControl.term(sc);
            throw new HoldCodeException();
        }
        else if(args[0].toString().equals("all"))
        {
            ScriptControl.termScripts();
            throw new HoldCodeException();
        }
        else
        {
            int i = getInt(args[0]);
            ScriptControl.term(ScriptControl.getScript(i));
            if(sc.getId() == i)
            {
                throw new HoldCodeException();
            }
        }
    }
    
    private static void ifFunction(Object[] args, Script sc)
    {
        if(Arrays.stream(args).anyMatch(s -> (boolean) s == false))
        {
            sc.gotoSpecialJumpLine();
            sc.jumpNextIfElse();
        }
    }  
    
    private static void whileFunction(Object[] args, Script sc)
    {
        if(Arrays.stream(args).anyMatch(s -> (boolean) s == false))
        {
            sc.gotoSpecialJumpLine(0);
        }
    } 
                  
    @SuppressWarnings(value = "unchecked")
    private static void sortList(List<Object> args, Script sc) 
    {
        Collections.sort(args, (Object o1, Object o2) -> ((Comparable) o1).compareTo(o2));
    }
    
    private static void scheduleGoto(Object[] args, Script sc)
    {
        SnuviScheduler.doScheduledTask(() -> 
        {
            if(!sc.isValid())
            {
                return;
            }
            try
            {
                sc.gotoLabel(args[1].toString());
                sc.setHalt(false);
                sc.runCode();
            }
            catch(Exception ex)
            {
                printQuestException(sc, ex, "(Scheduled Goto)");
            }
        }, getInt(args[0]));
    }
    
    private static void waitFor(Object[] args, Script sc) throws UnsupportedOperationException
    {           
        sc.resetLoopCounter();
        int i = getInt(args[0]);
        if(i < 1)
        {
            throw new UnsupportedOperationException();
        }
        sc.setHalt(true);
        SnuviScheduler.doScheduledTask(() -> 
        {                   
            if(sc == null || !sc.isValid())
            {
                return;
            }
            sc.setHalt(false);
            sc.runCode();
        }, i); 
    }   
    
    private static Number increaseVar(Object var, Script sc, int value)
    {
        double n = ((double) sc.getVar(var.toString())) + value;
        sc.setVar(var.toString(), n);
        return n;
    }   
    
    private static boolean isEqual(Object[] args)
    {
        if(args[0] == null)
        {
            return args[1] == null;
        }
        else if(args[1] == null)
        {
            return args[0] == null;
        }
        else if(args[1] instanceof Number && args[0] instanceof Number)
        {
            return ((Number) args[0]).doubleValue() == ((Number) args[1]).doubleValue();
        }
        return args[0].equals(args[1]);
    }
    
    private static void split(Object[] args, Script sc)
    {
        String[] parts = ScriptUtils.connect(args, 2).split(args[1].toString());
        ArrayList<Object> list = new ArrayList<>();
        for(String s : parts)
        {
            list.add(Code.convertInput(s));
        }
        sc.setVar(args[0].toString(), list);
    }   
    
    private static void print(Object[] args, Script sc)
    {
        if(args.length == 0)
        {
            ScriptUtils.print("");
        } 
        else if(args[0] instanceof Color)
        {
            ConsoleUtils.sendMessage(ScriptUtils.connect(args, 1), (Color) args[0]);
        }
        else
        {
            ScriptUtils.print(ScriptUtils.connect(args, 0));
        }
    }
    
    // -------------------------------------------------------------------------    
    // Read-Handler
    // ------------------------------------------------------------------------- 
    
    private static Color getColor(Object[] args, Script sc)
    {
        int r = getInt(args[0]);
        int g = getInt(args[1]);
        int b = getInt(args[2]);
        int a;
        if(args.length >= 4)
        {
            a = getInt(args[3]);
        }
        else
        {
            a = 255;
        }
        return new Color(r, g, b, a);
    }
    
    // -------------------------------------------------------------------------    
    // Graphics
    // ------------------------------------------------------------------------- 
    
    private static void newLine(Object[] args, Script sc)
    {
        sc.getGraphicPanel().addGraphic(getInt(args[0]), new Line((Color) args[1], getInt(args[2]), getInt(args[3]), getInt(args[4]), getInt(args[5])));
    }
    
    private static void newPoint(Object[] args, Script sc)
    {
        sc.getGraphicPanel().addGraphic(getInt(args[0]), new Point((Color) args[1], getInt(args[2]), getInt(args[3])));
    }
    
    private static void newOval(Object[] args, Script sc)
    {
        Oval oval = new Oval((Color) args[1], getInt(args[2]), getInt(args[3]), getInt(args[4]), getInt(args[5]));
        if(args.length >= 7)
        {
            oval.setFilled((boolean) args[6]);
        }
        sc.getGraphicPanel().addGraphic(getInt(args[0]), oval);
    }
    
    private static void newRectangle(Object[] args, Script sc)
    {
        Rectangle rec = new Rectangle((Color) args[1], getInt(args[2]), getInt(args[3]), getInt(args[4]), getInt(args[5]));
        if(args.length >= 7)
        {
            rec.setFilled((boolean) args[6]);
        }
        sc.getGraphicPanel().addGraphic(getInt(args[0]), rec);
    }
    
    private static void test(Object[] args, Script sc)
    {
        //sc.getGraphicPanel().objects.clear();
        //sc.getGraphicPanel().getGraphics().drawLine(20, 20, 30, 30);
    }
    
    // -------------------------------------------------------------------------    
    // Int-Handler, weil die Objekte double immer zu Double konvertieren
    // ------------------------------------------------------------------------- 
    
    private static int getInt(Object o)
    {
        return ((Number) o).intValue();
    }
    
    // -------------------------------------------------------------------------    
    // Zeit-Handler
    // ------------------------------------------------------------------------- 
    
    private static long getNextDay(Object[] args)       
    {
        GregorianCalendar cal = GregorianCalendar.from(ZonedDateTime.now());
        cal.setTimeInMillis((long) args[0]);
        cal.add(Calendar.DAY_OF_YEAR, 1);
        cal.set(Calendar.HOUR, 0);
        cal.set(Calendar.SECOND, 0);
        cal.set(Calendar.MINUTE, 0);
        cal.set(Calendar.MILLISECOND, 0);
        return cal.getTimeInMillis();   
    } 
    
    private static int getYear(Object[] args)       
    {
        GregorianCalendar cal = GregorianCalendar.from(ZonedDateTime.now());
        cal.setTimeInMillis((long) args[0]);
        return cal.get(Calendar.YEAR);   
    }
    
    private static int getMonth(Object[] args)       
    {
        GregorianCalendar cal = GregorianCalendar.from(ZonedDateTime.now());
        cal.setTimeInMillis((long) args[0]);
        return cal.get(Calendar.MONTH) + 1;   
    }
    
    private static int getDay(Object[] args)       
    {
        GregorianCalendar cal = GregorianCalendar.from(ZonedDateTime.now());
        cal.setTimeInMillis((long) args[0]);
        return cal.get(Calendar.DAY_OF_MONTH);   
    }
    
    private static int getHour(Object[] args)       
    {
        GregorianCalendar cal = GregorianCalendar.from(ZonedDateTime.now());
        cal.setTimeInMillis((long) args[0]);
        return cal.get(Calendar.HOUR_OF_DAY);   
    }
    
    private static int getMinute(Object[] args)       
    {
        GregorianCalendar cal = GregorianCalendar.from(ZonedDateTime.now());
        cal.setTimeInMillis((long) args[0]);
        return cal.get(Calendar.MINUTE);   
    }
    
    private static int getSecond(Object[] args)       
    {
        GregorianCalendar cal = GregorianCalendar.from(ZonedDateTime.now());
        cal.setTimeInMillis((long) args[0]);
        return cal.get(Calendar.SECOND);   
    }
}