package me.hammerle.code;

import java.awt.Graphics;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Stack;
import me.hammerle.console.ConsoleUtils;
import me.hammerle.exceptions.CodeTooLongException;
import me.hammerle.exceptions.GotoLabelNotFoundException;
import me.hammerle.exceptions.HoldCodeException;
import me.hammerle.graphics.GraphicFrame;
import me.hammerle.graphics.GraphicPanel;

public class Script 
{
    private final int id;
    private final String name;  

    private final HashMap<String, Object> variables;    
    private final HashMap<String, Integer> gotos;
    private final HashSet<String> events;
    private final Stack<Object> valueStack;
    private final Stack<Integer> returnStack;
    
    private final Code[] code;
    private int position;

    private int loopCounter;
    
    private boolean halt;
    private boolean valid;
    
    private int tryJumpLine;
    
    private GraphicFrame frame;
    private GraphicPanel panel;
    
    @SuppressWarnings("LeakingThisInConstructor")
    public Script(int id, String name, String code)
    {       
        this.id = id;
        this.name = name;
        
        variables = new HashMap<>();
        gotos = new HashMap<>();
        events = new HashSet<>();
        valueStack = new Stack<>();
        returnStack = new Stack<>();
        
        this.code = Code.generate(code, gotos);
        if(this.code == null)
        {
            ScriptControl.term(this);
            return;
        }
        
        position = 0;

        loopCounter = 0;
        
        halt = false;
        valid = true;
        
        tryJumpLine = -1;
        
        frame = null;
    }
    
    // -----------------------------------------------------------------------------------
    // Script-Daten
    // -----------------------------------------------------------------------------------
    
    public int getId() 
    {
        return id;
    }

    public String getName() 
    {
        return name;
    }
    
    public boolean isHalt() 
    {
        return halt;
    }
    
    public void setHalt(boolean b) 
    {
        halt = b;
    }
    
    public boolean isValid() 
    {
        return valid;
    }
    
    public void setInvalid() 
    {
        valid = false;
    }
    
    public int getActiveCodeLine()
    {
        return position;
    }
    
    // -----------------------------------------------------------------------------------
    // Event-Handling
    // -----------------------------------------------------------------------------------
    
    public void loadEvent(String s)
    {
        events.add(s);
    }
    
    public boolean isLoadedEvent(String s)
    {
        return events.contains(s);
    }
    
    public void unloadEvent(String s)
    {
        events.remove(s);
    }
    
    // -----------------------------------------------------------------------------------
    // Script-Flow
    // -----------------------------------------------------------------------------------
    
    public void runCode() 
    {
        if(this.isValid())
        {
            try
            {
                while(position < code.length)
                {
                    code[position].executeFunction(this, valueStack);
                    position++;
                    if(isHalt())
                    {
                        ConsoleUtils.scrollToBottom();
                        return;
                    }
                }
                ConsoleUtils.scrollToBottom();
                ScriptControl.term(this);
            }
            catch(Exception ex)
            {
                if(ex.getClass() != HoldCodeException.class)
                {
                    CodeParser.printQuestException(this, ex, code[position].toString());
                }
                position++;
                ConsoleUtils.scrollToBottom();
            }
        }
    }
    
    // -----------------------------------------------------------------------------------
    // Variablen
    // -----------------------------------------------------------------------------------
    
    public void setVar(String var, Object value)
    {
        variables.put(var, value);   
    }
    
    public Object getVar(String var)
    {
        return variables.get(var);
    }
    
    public void removeVar(String var)
    {
        variables.remove(var);
    }
    
    // -----------------------------------------------------------------------------------
    // Goto, Return
    // -----------------------------------------------------------------------------------
    
    public void incLoopCounter()
    {
        loopCounter++;
        if(loopCounter > 50)
        {
            resetLoopCounter();
            throw new CodeTooLongException();
        }
    }
    
    public void gotoLabel(String label)
    {
        try
        {
            int i = gotos.get(label);
            incLoopCounter();
            position = i - 1;
        }
        catch(NullPointerException ex)
        {
            throw new GotoLabelNotFoundException(label);
        }
    }  
    
    public void resetLoopCounter()
    {
        loopCounter = 0;
    }
    
    public void gotoLabelWithReturn(String label)
    {
        returnStack.push(position);
        gotoLabel(label);
    }
    
    public void doReturn()
    {
        position = returnStack.pop(); 
    }
    
    public void gotoSpecialJumpLine(int i)
    {
        position += code[position].getJumpLine() + i;
    }
    
    public void gotoSpecialJumpLine()
    {
        gotoSpecialJumpLine(-1);
    }
    
    public void jumpNextIfElse()
    {
        if(code.length <= position + 1)
        {
            return;
        }
        String s = code[position + 1].getFunction();
        if(s == null || !s.equals("else"))
        {
            return;
        }
        position++;
    }
    
    // -----------------------------------------------------------------------------------
    // try
    // -----------------------------------------------------------------------------------
    
    public void gotoTryJumpLine()
    {
        position = tryJumpLine; 
    }
    
    public void saveTryJumpLine()
    {
        tryJumpLine = position + code[position].getJumpLine();
    }
    
    public void resetTryJumpLine()
    {
        tryJumpLine = -1;
    }
    
    public int getTryJumpLine()
    {
        return tryJumpLine; 
    }
    
    // -----------------------------------------------------------------------------------
    // Frame
    // -----------------------------------------------------------------------------------
    
    public void newFrame(int width, int height)
    {
        if(frame == null)
        {
            frame = new GraphicFrame(width, height, this);
            panel = frame.graphics;
        }
    }
    
    public GraphicPanel getGraphicPanel()
    {
        return panel;
    }
    
    public void closeFrame()
    {
        if(frame != null)
        {
            frame.dispose();
        }
    }
}
